// interface to manage socket communications
//
#include <iostream>
#include <string>

#include <sys/types.h>  
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ifaddrs.h>
#include <unistd.h>
#include <netdb.h>
#include <string.h>
#include <errno.h>

using namespace std;

#define MHN 256

class ReceiveChannel {

private:
  int socks, sockc, socka; 
  int ports, portc; 
  string name; 
  int sp = 12345;
  int backlog = 5; 
  char hostname[MHN];
  struct in_addr local_addr;
  struct sockaddr_in sas,sac;
  
public:

  ReceiveChannel(string n) {
    name = n;

    // server socket, therefore bind to local IP address
    socks = socket(AF_INET, SOCK_STREAM, 0);
    if(socks == -1) {
      perror("While opening the server socket: "); 
      return;
    }
    // get local host name 
    if(gethostname(hostname, MHN)!=0) {
      perror("While getting host name:"); 
      return;
    }
    // prepare address 
    localAddr();
    sas.sin_family = AF_INET;
    sas.sin_port = htons(sp);
    memcpy(&(sas.sin_addr),&local_addr, sizeof(local_addr));
    
    int ret;
    while((ret = bind(socks, (sockaddr *)&sas, sizeof(sas))) != 0) {
      sas.sin_port = htons(++sp);
    }

    ret = listen(socks, backlog);
    if(ret == -1) {
      perror("While listening:"); 
      return;
    }
    cout << "Published ss at " << inet_ntoa(local_addr) << " on port " << sp << endl;
    
    return;
  }

  bool StartOperation() {
    struct sockaddr_in a;
    socklen_t al; 

    socka = accept(socks,(struct sockaddr *)&a, &al);
    if(socka == -1) {
      cerr << "Error accepting " << endl;
      return(false);
    }
    
    return(true);
  }

  void localAddr() {
    struct ifaddrs * ifAddrStruct=NULL;
    struct ifaddrs * ifa=NULL;
    void * tmpAddrPtr=NULL;
    
    getifaddrs(&ifAddrStruct);
    
    for (ifa = ifAddrStruct; ifa != NULL; ifa = ifa->ifa_next) {
      if (!ifa->ifa_addr) {
	continue;
      }
      if (ifa->ifa_addr->sa_family == AF_INET) { // check it is IP4
	// is a valid IP4 Address
	tmpAddrPtr=&((struct sockaddr_in *)ifa->ifa_addr)->sin_addr;
	char addressBuffer[INET_ADDRSTRLEN];
	inet_ntop(AF_INET, tmpAddrPtr, addressBuffer, INET_ADDRSTRLEN);
	string slo("lo");
	if(!(slo == ifa->ifa_name)) {
	  printf("%s IP Address %s\n", ifa->ifa_name, addressBuffer);
	  inet_aton(addressBuffer,&local_addr);
	  return;
	}
      } 
    }
    if (ifAddrStruct!=NULL) freeifaddrs(ifAddrStruct);
    return;
  }

  // send message to the connected socket (retro-action)
  void send(char * buffer, int len) {
    int ret = write(socka, buffer, len);
    if(ret != len) {
      perror("While writing:");
    }
    return;
  }

  // receive message from the connected socket
  void receive(char * buffer, int maxlen) {
    // want to receive exactly that amount of chars
    int received = 0;
    while(received != maxlen) {
      int ret = read(socka,&(buffer[received]),maxlen-received);
      if(ret < 0) {
	perror("While reading: ");
      } else {
	received += ret;
      }
    }
    return; 
  }

};


class SendChannel {

 private:
  int socks, sockc, socka; 
  int ports, portc; 
  string name; 
  int sp = 12345;
  int backlog = 5; 
  char hostname[MHN];
  struct in_addr local_addr;
  struct sockaddr_in sas,sac;
  
public:
  
  SendChannel(string ip, int port) {
    sockc = socket(AF_INET, SOCK_STREAM, 0);
    if(sockc == -1) {
      perror("While opening the socket: ");
      return;
    }
    struct in_addr addr;
    inet_aton(ip.c_str(),&addr);
    sac.sin_family = AF_INET;
    sac.sin_port = htons(port);
    memcpy(&(sac.sin_addr), &addr, sizeof(addr));

    int ret = connect(sockc, (struct sockaddr *) &sac, sizeof(sac));
    while(ret == -1) {
      perror("While connecting: ");
      sleep(1);
      ret = connect(sockc, (struct sockaddr *) &sac, sizeof(sac));
    }
    cout << "Socket connected (" << ip << "," << port << ")" << endl;
    
    return;
  }

    // send message to the connected server socket
  void send(char * buffer, int len) {
    int ret = write(sockc, buffer, len);
    if(ret != len) {
      perror("While writing:");
    }
    return;
  }

  // receive message from the connected socket (retro-action)
  void receive(char * buffer, int maxlen) {
    // want to receive exactly that amount of chars
    int received = 0;
    while(received != maxlen) {
      int ret = read(sockc,&(buffer[received]),maxlen-received);
      if(ret < 0) {
	perror("While reading: ");
      } else {
	received += ret;
      }
    }
    return; 
  }

  void EndOperation() {
    close(sockc);
  }
  
}; 


